#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include "xwin.h"
#include "ftread.h"

static int is_fs;
static char path_sep;

typedef struct FileEnt {
  int isdir;
  char size[8];
  char *name;
};

static FileEnt *ents;
static int nents=0, maxents=0, firstent;
static char path[1024];
static char *suffix;
static int path_offset_x = 0;

static int em=0, ems, fw;

static void
new_info(int clear)
{
  if (em == 0 || clear > 1)
    {
      em = xfs->ascent + xfs->descent;
      ems = (win_height-4) / em - 1;
      fw = xfs->max_bounds.width;
      XSetForeground(display, gc, grays[0]);
      XFillRectangle(display, window, gc, 0, 0, win_width, win_height);
    }
  else if (clear)
    {
      XSetForeground(display, gc, grays[0]);
      XFillRectangle(display, window, gc, 0, 0, win_width, 2+em);
    }
  XSetForeground(display, gc, grays[15]);

  int pw = strlen(path);
  path_offset_x = win_width - pw * fw - 4;
  if (path_offset_x > 0)
    path_offset_x = 0;
  XDrawImageString(display, window, gc, 2+path_offset_x, 2+xfs->ascent, path, strlen(path));

  if (nents <= 0)
    return;

  for (int i=0; i<ems; i++)
    {
      if (i+firstent >= nents)
	break;
      int x, y = 2 + em + i*em + xfs->ascent;
      int sl = strlen(ents[i+firstent].name);

      XSetForeground(display, gc, grays[15]);

      if (is_fs)
	{
	  XDrawImageString(display, window, gc, 2 + em, y,
			   ents[i+firstent].size, strlen(ents[i+firstent].size));
	  XDrawImageString(display, window, gc, 2 + em + 6*fw, y,
			   ents[i+firstent].name, sl);
	  x = 2+em+6*fw+fw*strlen(ents[i+firstent].name);
	}
      else
	{
	  XDrawImageString(display, window, gc, 2 + em, y,
			   ents[i+firstent].name, sl);
	  x = 2+em+fw*strlen(ents[i+firstent].name);
	}

      if (clear)
	{
	  XSetForeground(display, gc, grays[0]);
	  XFillRectangle(display, window, gc,
			 x, y-xfs->ascent, win_width-x, x+xfs->descent);
	}
    }

  int sbheight = em * ems;
  int sbtop = firstent * sbheight / nents + 1 + 2+em;
  int sbbot = (firstent+ems) * sbheight / nents - 1 + 2+em;
  if (sbbot > sbheight-1+2+em)
    sbbot = sbheight-1+2+em;

  int ts = grays[8];
  int bs = grays[15];
  int trough = grays[4];
  int button = grays[0];

  XSetForeground(display, gc, ts);
  XDrawLine (display, window, gc, 2, 2+em+sbheight, em, 2+em+sbheight);
  XDrawLine (display, window, gc, em, 2+em+1, em, 2+em+sbheight);

  XDrawLine (display, window, gc, 2, sbtop, em-1, sbtop);
  XDrawLine (display, window, gc, 2, sbtop, 2, sbbot);

  XSetForeground(display, gc, bs);
  XDrawLine (display, window, gc, 1, 2+em, em, 2+em);
  XDrawLine (display, window, gc, 1, 2+em, 1, 2+em+sbheight);

  XDrawLine (display, window, gc, 3, sbbot, em-1, sbbot);
  XDrawLine (display, window, gc, em-1, sbtop, em-1, sbbot);

  XSetForeground(display, gc, trough);
  if (firstent > 0)
    XFillRectangle (display, window, gc, 2, 2+em+1, em-2, sbtop-(2+em+1));
  if (firstent+ems < nents)
    XFillRectangle (display, window, gc, 2, sbbot+1, em-2, sbheight+1+em-sbbot);

  XSetForeground(display, gc, button);
  XFillRectangle (display, window, gc, 3, sbtop+1, em-4, sbbot-sbtop-1);
}

static int
ent_sort (const void *va, const void *vb)
{
  FileEnt *a = (FileEnt *)va;
  FileEnt *b = (FileEnt *)vb;
  if (a->isdir != b->isdir)
    return b->isdir - a->isdir;
  return strcmp (a->name, b->name);
}

static void
scan_dir()
{
  DIR *dir = opendir(path);
  struct dirent *de;

  char *endp = path + strlen(path);
  *endp++ = path_sep;

  while ((de = readdir(dir)) != 0)
    {
      if (de->d_name[0] == '.')
	continue;
      if (strchr(de->d_name, '~'))
	continue;
      if (strchr(de->d_name, '#'))
	continue;
      if (nents >= maxents)
	{
	  maxents += 20;
	  ents = (FileEnt *)realloc (ents, maxents * sizeof(FileEnt));
	}
      struct stat s;
      strcpy(endp, de->d_name);
      stat(path, &s);

      if (S_ISREG(s.st_mode)
	  && suffix
	  && strcmp (suffix, de->d_name + strlen(de->d_name) - strlen(suffix)) != 0)
	continue;

      if (S_ISDIR(s.st_mode))
	{
	  ents[nents].isdir = 1;
	  strcpy(ents[nents].size, "<dir>");
	}
      else if (S_ISREG(s.st_mode))
	{
	  ents[nents].isdir = 0;
	  int suf = ' ';
	  int byt = s.st_size;
	  if (s.st_size > 10 * 1024 * 1024)
	    suf = 'm', byt /= 1024*1024;
	  else if (s.st_size > 10 * 1024)
	    suf = 'k', byt /= 1024;
	  sprintf(ents[nents].size, "%4d%c", byt, suf);
	}
      else
	{
	  continue;
	}
      ents[nents].name = strdup(de->d_name);
      nents++;
    }

  *--endp = 0;

}

static void
scan_font()
{
  static char **all_fonts;
  static int nfonts = 0;
  if (nfonts == 0)
    {
      int max = 10;
      while (1) {
	all_fonts = XListFonts(display, "*", max, &nfonts);
	if (nfonts < max)
	  break;
	XFreeFontNames(all_fonts);
	max *= 10;
      };
    }

  int i, j;
  int pathlen = strlen(path);
  for (i=0; i<nfonts; i++)
    {
      if (memcmp(all_fonts[i], path, pathlen))
	continue;
      if (pathlen && all_fonts[i][pathlen] != '-')
	continue;
      char *endp = all_fonts[i] + pathlen;
      char *begp = endp;
      if (*path)
	begp++, endp++; /* skip '-' */
      while (*endp == '-')
	endp++;
      while (*endp && *endp != '-')
	endp++;
      char endc = *endp;
      if (endp && strlen(endp) > 3)
	*endp = 0;

      for (j=nents-1; j>=0; j--)
	if (strcmp(ents[j].name, begp) == 0)
	  break;
      if (j == -1)
	{
	  if (nents >= maxents)
	    {
	      maxents += 20;
	      ents = (FileEnt *)realloc (ents, maxents * sizeof(FileEnt));
	    }
	  ents[nents].isdir = endc ? 1 : 0;
	  ents[nents].name = strdup (begp);
	  nents++;
	}
      *endp = endc;
    }
}

static void
scan()
{
  firstent = 0;
  while (nents > 0)
    {
      nents--;
      free (ents[nents].name);
    }
  if (is_fs)
    scan_dir();
  else
    scan_font();

  qsort(ents, nents, sizeof(ents[0]), ent_sort);
  new_info(2);
}

static int scroll_y=0, scroll_initial;

static void
scroll_to(int y)
{
  if (scroll_y == 0)
    return;
  int dy = y - scroll_y;
  int oldfirstent = firstent;
  firstent = scroll_initial + dy * nents / (em*ems);
  if (firstent > nents-ems)
    firstent = nents-ems;
  if (firstent < 0)
    firstent = 0;
  if (firstent != oldfirstent)
    new_info(1);
}

void
click_on_path(int x)
{
  int idx = (x-2-path_offset_x) / fw;
  if (idx >= strlen(path))
    return;
  if (idx < 0)
    idx = 0;
  if (is_fs)
    {
      while (path[idx] && path[idx] != path_sep)
	idx++;
      if (idx == 0)
	idx++;
      if (path[idx] == 0)
	return;
    }
  else
    {
      while (idx>0 && path[idx] != path_sep)
	idx--;
    }
  path[idx] = 0;
  scan();
}

static int done = 0;

void
click_on_file(int y)
{
  int e = (y-2-em) / em + firstent;
  if (e < 0 || e >= nents)
    return;
  char *endp = path + strlen(path);
  if (endp != path && endp[-1] != path_sep)
    *endp++ = path_sep;
  strcpy(endp, ents[e].name);
  if (ents[e].isdir)
    {
      scan();
      while (!is_fs && nents == 1)
	{
	  strcat(path, "-");
	  strcat(path, ents[0].name);
	  scan();
	}
      if (nents == 0)
	done = 1;
    }
  else
    done = 1;

  if (done && !is_fs)
    {
      char *dash = path;
      int i;
      font_is_scalable = 0;
      for (i=0; i<7; i++)
	{
	  dash = strchr(dash, '-');
	  if (!dash)
	    break;
	  dash++;
	}
      if (dash && memcmp (dash, "0-", 2) == 0)
	{
	  memmove (dash+1, dash, strlen(dash)+1);
	  *dash++ = '%';
	  *dash++ = 'd';
	  font_is_scalable = 1;
	}
    }
}

char *
common_selector()
{
  if (maxents == 0)
    {
      maxents = 10;
      ents = (FileEnt *)malloc(maxents * sizeof(FileEnt));
    }
  nents = 0;

  scan();
  done = 0;

  XEvent e;
  int i;
  while (!done)
    {
      xwin_event(&e);
      switch (e.type)
	{
	case Expose:
	  new_info(0);
	  break;

	case ButtonPress:
	  scroll_y = 0;
	  if (e.xbutton.y < em+2)
	    click_on_path(e.xbutton.x);
	  else if (e.xbutton.x > em+2)
	    {
	      click_on_file(e.xbutton.y);
	    }
	  else
	    {
	      scroll_y = e.xbutton.y;
	      scroll_initial = firstent;
	    }
	  break;
	  
	case ButtonRelease:
	  scroll_to(e.xbutton.y);
	  break;
		    
	case MotionNotify:
	  scroll_to(e.xmotion.y);
	  break;
	}
    }

  return path;
}

char *
file_selector(char *suffixP)
{
  path_sep = '/';
  is_fs = 1;
  suffix = suffixP;
  getcwd(path, sizeof(path));
  return common_selector();
}

char *
font_selector()
{
  path_sep = '-';
  is_fs = 0;
  suffix = 0;
  strcpy(path, "");
  return common_selector();
}
